#include "graph.h"

// Initial size of lookup graph. It will double each time capacity is reached.
#define LOOKUP_GR_INIT_SIZE 64

// Assume VERY coarsly that the number of unique terms will be in general
// 1.7 times the number of triples. This is conservative to maintain load
// factor low.
#define IDX_SIZE_RATIO 1.7


typedef enum KSetFlag {
    LSUP_KS_NONE        = 0,
    LSUP_KS_CHECK_CAP   = 1 << 0,
    LSUP_KS_CHECK_DUP   = 1 << 1,
} KSetFlag;

enum {
    HS_ERROR        = -1,
    HS_PRESENT      = 0,
    HS_EMPTY        = 1,
    HS_TOMBSTONE    = 2,
};

/**
 * Index (identity) hashing function.
 *
 * Since the key is already a strong hash, reuse it for bucket allocation.
 */
static inline khint64_t idx_key_hash_fn(LSUP_Key key)
{ return key; }

/**
 * Triple Key hash.
 *
 * Since triple keys are already hashes, interlace the first bytes of each
 * key element to preserve the individual identities.
 */
static inline khint64_t tkey_hash_fn(LSUP_TripleKey spok)
{
    khint64_t rc = (khint64_t)XXH64(spok, sizeof(LSUP_TripleKey), SEED);
    TRACE("Key hash: 0x%lx\n", rc);
    return rc;
}

static inline int tkey_eq_fn(LSUP_Key *a, LSUP_Key *b)
{
    TRACE("a addr: %p", a);
    TRACE("b addr: %p", b);
    TRACE("Comparing a: {%lx, %lx, %lx}", a[0], a[1], a[2]);
    TRACE("Wtih      b: {%lx, %lx, %lx}", b[0], b[1], b[2]);
    // Evaluate from the least to the most probable to match.
    return (
            a[2] == b[2] &&
            a[0] == b[0] &&
            a[1] == b[1]);
}

KHASH_INIT(Keys, LSUP_Key *, char, 0, tkey_hash_fn, tkey_eq_fn)

KHASH_INIT(
        Index, LSUP_Key, LSUP_Buffer, 1, idx_key_hash_fn, kh_int_hash_equal)

typedef struct Graph {
    LSUP_store_type store_type;     // In-memory or MDB-backed
    LSUP_Term *uri;                 // Graph "name" (URI)
    khash_t(Keys) *keys;
    khash_t(Index) *idx;            // Dictionary of keys to serialized terms
} Graph;

/**
 * Extern inline functions.
 */
size_t LSUP_graph_size(LSUP_Graph *gr);

size_t LSUP_graph_capacity(LSUP_Graph *gr);


/**
 * Callback type for key comparison.
 */
typedef bool (*LSUP_key_cmp_fn_t)(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2);


// Dummy callback for queries with all parameters unbound. Returns true.
static bool lookup_none_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return true; }

//  Keyset lookup for S key.
static bool lookup_sk_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][0] == k1; }

//  Keyset lookup for P key.
static bool lookup_pk_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][1] == k1; }

//  Keyset lookup for O key.
static bool lookup_ok_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][2] == k1; }

//  Keyset lookup for S and P keys.
static bool lookup_skpk_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][0] == k1 && spok[0][1] == k2; }

//  Keyset lookup for S and O keys.
static bool lookup_skok_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][0] == k1 && spok[0][2] == k2; }

//  Keyset lookup for P and O keys.
static bool lookup_pkok_cmp_fn(
        const LSUP_TripleKey* spok, const LSUP_Key k1, const LSUP_Key k2)
{ return spok[0][1] == k1 && spok[0][2] == k2; }


/* * * Post-lookup callbacks * * */

int match_add_fn(LSUP_Graph *src, LSUP_Graph *dest, khiter_t cur, void *ctx);

int match_rm_fn(LSUP_Graph *src, LSUP_Graph *dest, khiter_t cur, void *ctx);


/* * * KEYSETS * * */

static inline bool is_null_trp(const LSUP_TripleKey *trp)
{
    return (
            *trp[0] == NULL_KEY &&
            *trp[1] == NULL_KEY &&
            *trp[2] == NULL_KEY);
}


/* * * GRAPH * * */

int
LSUP_graph_init(
        LSUP_Graph *gr, size_t capacity, char *uri_str,
        LSUP_store_type store_type)
{
    if (uri_str == NULL) {
        gr->uri = LSUP_term_new(
                LSUP_TERM_URI, LSUP_term_gen_random_str(), NULL, NULL);
    } else {
        gr->uri = LSUP_term_new(LSUP_TERM_URI, uri_str, NULL, NULL);
    }

    gr->keys = kh_init_Keys();
    kh_resize_Keys(gr->keys, capacity);

    switch (store_type ) {
        case LSUP_STORE_MEM:
            gr->idx = kh_init_Index();
            kh_resize_Index(gr->idx, capacity * IDX_SIZE_RATIO);
            break;

        case LSUP_STORE_MDB:
            // TODO

        default:
            return -1;
    }

    return LSUP_OK;
}


LSUP_Graph *
LSUP_graph_new(size_t capacity, char *uri_str, LSUP_store_type store_type)
{
    LSUP_Graph *gr;
    CRITICAL(gr = malloc(sizeof(LSUP_Graph)));

    LSUP_graph_init(gr, capacity, uri_str, store_type);

    return gr;
}


/**
 * Copy triples from a source graph into a destination one.
 *
 * The destination graph is not initialized, so the copy is cumulative.
 */
static int graph_copy_contents(LSUP_Graph *src, LSUP_Graph *dest)
{
    LSUP_Triple trp;
    trp.s = NULL;
    trp.p = NULL;
    trp.o = NULL;

    return LSUP_graph_match_callback(
            src, dest, &trp, match_add_fn, true, NULL);
}


int
LSUP_graph_copy(LSUP_Graph *dest, LSUP_Graph *src)
{
    LSUP_graph_init(dest, LSUP_graph_size(src), NULL, src->store_type);

    return graph_copy_contents(src, dest);
}


int
LSUP_graph_resize(LSUP_Graph *gr, size_t size)
{
    kh_resize_Keys(gr->keys, size);
    kh_resize_Index(gr->idx, size * IDX_SIZE_RATIO);

    return LSUP_OK;
}


size_t
LSUP_graph_capacity(LSUP_Graph *gr) { return kh_end(gr->keys); }


char *
LSUP_graph_uri(LSUP_Graph *gr) { return gr->uri->data; }


size_t
LSUP_graph_size(LSUP_Graph *gr) { return kh_size(gr->keys); }


int
LSUP_graph_add_triple(LSUP_Graph *gr, const LSUP_Triple *spo)
{
    LSUP_SerTerm sspo[3];

    LSUP_term_serialize(spo->s, sspo);
    LSUP_term_serialize(spo->p, sspo + 1);
    LSUP_term_serialize(spo->o, sspo + 2);

    LSUP_TripleKey spok = NULL_TRP;

    // Add term to index.
    int status;
    khiter_t cur;
    for (int i = 0; i < 3; i++) {
        spok[i] = LSUP_sterm_to_key(sspo + i);

        cur = kh_put_Index(gr->idx, spok[i], &status);
        if (status == HS_EMPTY || status == HS_TOMBSTONE) {
            // If term is not indexed, store the struct in the index.
            kh_value(gr->idx, cur) = sspo[i];
        } else {
            // If term is already in the index, discard and free it.
            LSUP_buffer_done(sspo + i);
        }
    }

    // Add triple.
    TRACE("Inserting spok: {%lx, %lx, %lx}", spok[0], spok[1], spok[2]);
    cur = kh_put_Keys(gr->keys, spok, &status);
    //if (status == HS_EMPTY || status == HS_TOMBSTONE)
    //    kh_key(gr->keys, cur) = spok;
    TRACE("cur: %d\n", cur);
    TRACE("Keyset size: %d\n", kh_size(gr->keys));
    TRACE("Insert status: %d\n", status);

    return status == HS_PRESENT ? LSUP_NOACTION : LSUP_OK;
}


int
LSUP_graph_add(LSUP_Graph *gr, const LSUP_Triple data[], size_t data_size)
{
    // TODO Decouple this and build interface for memory and MDB integration.

    // Resize all at once if needed.
    if (LSUP_graph_capacity(gr) < LSUP_graph_size(gr) + data_size)
        LSUP_graph_resize(gr, LSUP_graph_size(gr) + data_size);

    int rc = LSUP_NOACTION;
    for (size_t i = 0; i < data_size; i++) {
        TRACE("Inserting triple #%lu\n", i);
        if (LIKELY(LSUP_graph_add_triple(gr, data + i) == LSUP_OK))
            rc = LSUP_OK;
    }

    return rc;
}


bool
LSUP_graph_contains(const LSUP_Graph *gr, const LSUP_Triple *spo)
{
    LSUP_Key sk = LSUP_term_to_key(spo->s);
    LSUP_Key pk = LSUP_term_to_key(spo->p);
    LSUP_Key ok = LSUP_term_to_key(spo->o);

    LSUP_TripleKey spok = {sk, pk, ok};

    return kh_get_Keys(gr->keys, spok) != kh_end(gr->keys);
}


int LSUP_graph_match_callback(
        LSUP_Graph *gr, LSUP_Graph *res, const LSUP_Triple *spo,
        keyset_match_fn_t callback_fn, bool match_cond, void *ctx)
{
    if (kh_size(gr->keys) == 0)
        return LSUP_NOACTION;

    LSUP_Key k1, k2;
    LSUP_key_cmp_fn_t cmp_fn;

    LSUP_Key sk = LSUP_term_to_key(spo->s);
    LSUP_Key pk = LSUP_term_to_key(spo->p);
    LSUP_Key ok = LSUP_term_to_key(spo->o);

    if (sk != NULL_KEY && pk != NULL_KEY && ok != NULL_KEY) {
        LSUP_TripleKey spok = {sk, pk, ok};

        if (match_cond == true) {
            // Shortcut for 3-term match—only if match_cond is true.
            LSUP_graph_init(res, 1, NULL, LSUP_STORE_MEM);
            khint_t cur = kh_get_Keys(gr->keys, spok);
            if(cur != kh_end(gr->keys)) {
                callback_fn(gr, res, cur, ctx);
                return LSUP_OK;
            } else {
                return LSUP_NOACTION;
            }
        } else {
            // For negative condition (i.e. "apply this function to all triples
            // except the matching one")
            int rc = LSUP_NOACTION;
            for (khiter_t i = kh_begin(gr->keys); i != kh_end(gr->keys); i++) {
                LSUP_Key *cur = kh_key(gr->keys, i);
                if (
                    cur[0] != spok[0] ||
                    cur[1] != spok[1] ||
                    cur[2] != spok[2]
                ) {
                    callback_fn(gr, res, i, ctx);
                    rc = LSUP_OK;
                }
            }

            return rc;
        }

    } else if (sk != NULL_KEY) {
        k1 = sk;

        if (pk != NULL_KEY) { // s p ?
            k2 = pk;
            cmp_fn = lookup_skpk_cmp_fn;

        } else if (ok != NULL_KEY) { // s ? o
            k2 = ok;
            cmp_fn = lookup_skok_cmp_fn;

        } else { // s ? ?
            cmp_fn = lookup_sk_cmp_fn;

        }

    } else if (pk != NULL_KEY) {
        k1 = pk;

        if (ok != NULL_KEY) { // ? p o
            k2 = ok;
            cmp_fn = lookup_pkok_cmp_fn;

        } else { // ? p ?
            cmp_fn = lookup_pk_cmp_fn;
        }

    } else if (ok != NULL_KEY) { // ? ? o
        k1 = ok;
        cmp_fn = lookup_ok_cmp_fn;

    } else {
        printf("WARNING: no bound terms, making a compact copy.\n");
        return LSUP_graph_copy(res, gr);
    }

    for(khiter_t i = kh_begin(gr->keys); i != kh_end(gr->keys); i++) {
        LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(gr->keys, i);
        if (cmp_fn(spok, k1, k2) == match_cond)
            callback_fn(gr, res, i, ctx);
    }

    return LSUP_OK;
}


int LSUP_graph_lookup(LSUP_Graph *gr, LSUP_Graph *res, const LSUP_Triple *spo)
{
    LSUP_graph_init(res, LOOKUP_GR_INIT_SIZE, NULL, LSUP_STORE_MEM);

    LSUP_graph_match_callback(gr, res, spo, &match_add_fn, true, NULL);

    return LSUP_OK;
}


int LSUP_graph_join(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_graph_copy(res, gr1);

    return graph_copy_contents(gr2, res);
}


int LSUP_graph_subtract(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    if (kh_size(gr2->keys) == 0) return LSUP_graph_copy(gr1, res);

    LSUP_graph_init(res, LSUP_graph_capacity(gr1), NULL, LSUP_STORE_MEM);

    if (kh_size(gr1->keys) == 0) return LSUP_OK;

    for(khiter_t i = kh_begin(gr1->keys); i != kh_end(gr1->keys); i++) {
        LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(gr1->keys, i);

        khiter_t cur = kh_get_Keys(gr2->keys, *spok);
        if (cur == kh_end(gr2->keys)) {
            match_add_fn(res, gr1, cur, NULL);
        }
    }

    return LSUP_OK;
}


int LSUP_graph_intersect(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_graph_init(res, LSUP_graph_capacity(gr1), NULL, LSUP_STORE_MEM);

    if (kh_size(gr1->keys) == 0 || kh_size(gr2->keys) == 0)
        return LSUP_OK;

    for(khiter_t i = kh_begin(gr1->keys); i != kh_end(gr1->keys); i++) {
        LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(gr1->keys, i);

        khiter_t cur = kh_get_Keys(gr2->keys, *spok);
        if (cur != kh_end(gr2->keys)) {
            match_add_fn(res, gr1, cur, NULL);
        }
    }

    return LSUP_OK;
}


int LSUP_graph_xor(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    if (kh_size(gr1->keys) == 0) return LSUP_graph_copy(gr2, res);
    if (kh_size(gr2->keys) == 0) return LSUP_graph_copy(gr1, res);

    LSUP_graph_init(
            res, min(LSUP_graph_capacity(gr1), LSUP_graph_capacity(gr2)),
            NULL, LSUP_STORE_MEM);

    for(khiter_t i = kh_begin(gr1->keys); i != kh_end(gr1->keys); i++) {
        LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(gr1->keys, i);

        khiter_t cur = kh_get_Keys(gr2->keys, *spok);
        if (cur == kh_end(gr2->keys)) {
            match_add_fn(res, gr1, cur, NULL);
        }
    }

    for(khiter_t i = kh_begin(gr2->keys); i != kh_end(gr2->keys); i++) {
        LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(gr2->keys, i);

        khiter_t cur = kh_get_Keys(gr1->keys, *spok);
        if (cur == kh_end(gr1->keys)) {
            match_add_fn(res, gr2, cur, NULL);
        }
    }

    return LSUP_OK;
}


void
LSUP_graph_free(LSUP_Graph *gr)
{
    if (LIKELY(gr != NULL)) {
        LSUP_term_free(gr->uri);

        // Free up triples.
        /*
        for(khiter_t i = kh_begin(gr->keys); i != kh_end(gr->keys); ++i) {
            if(kh_exist(gr->keys, i))
                free(&kh_value(gr->keys, i));
        }
        */
        kh_destroy_Keys(gr->keys);

        // Free up index entries and index.
        for(khiter_t i = kh_begin(gr->idx); i != kh_end(gr->idx); ++i) {
            if(kh_exist(gr->idx, i))
                LSUP_buffer_done(&kh_value(gr->idx, i));
        }
        kh_destroy_Index(gr->idx);

        free(gr);
    }
}


/* * CALLBACKS * */

/**
 * Callback for adding a matched triple.
 *
 * Adds the current triple in src to dest. No duplicate check.
 *
 * The source graph cursor must be set to the triple to be copied.
 */
int match_add_fn(LSUP_Graph *src, LSUP_Graph *dest, khiter_t cur, void *ctx)
{
    LSUP_TripleKey *spok = (LSUP_TripleKey*)&kh_key(src->keys, cur);
    // Add term to index.
    int status;
    for (int i = 0; i < 3; i++) {
        khiter_t dest_cur = kh_put_Index(dest->idx, *spok[i], &status);
        if (
                kh_exist(src->idx, *spok[i] &&
                (status == HS_EMPTY || status == HS_TOMBSTONE))) {
            // If term is not indexed, store the struct in the index.
            LSUP_buffer_copy(
                    &kh_value(dest->idx, dest_cur), &kh_value(src->idx, cur));
        }
    }

    // Add triple.
    kh_put_Keys(src->keys, *spok, &status);

    return LSUP_OK;
}


/**
 * Callback for removing a matched triple.
 */
int match_rm_fn(LSUP_Graph *src, LSUP_Graph *dest, khint_t cur, void *ctx)
{
    kh_del_Keys(dest->keys, cur);

    return LSUP_OK;
}
